/*----------------------------------------------------------------------------
;
; Generate byte pattern to stdout.  (c) 2019 PaulHoule.com
;
; Change Log
; ----------
; 20190914 ver 2.3: (1) small bug fixes; (2) command line fetched as wide
;	char, so @file argument(s) can now be unicode filenames; (3) implemented
;	\U and \u string escape sequence; (4) added UTF-16 "wide char" string
;	type via L"string", this generates surrogate pairs for \U >= 0x10000.
;
; 20190901 ver 2.1: allow double quoted string arg(s); also, decimal output
;	size arg can be omitted if first argument is a string or @{file}.
;
;---------------------------------------------------------------------------*/

#ifdef PH_								// if "tiny .exe" stdlib (optional)
#include "ph_.h"
#else									// else normal compiler stdlib
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdint.h>
#endif

typedef struct spbuf {					// stores data buffers in a list
	struct spbuf *pNext;				// next PBUF, or NULL if last
	size_t totprevbused;				// sum of all previous bused in list
	uint32_t  bbytes;					// bytes actually allocated to buf[]
	uint32_t  bused;					// bytes currently used in buf[]
	uint8_t buf[65536];					// (keep this member 8-byte aligned)
} PBUF;

typedef struct {						// for parsed command line args
	uint64_t fsize;						// parsed filesize (1st arg)
	enum { AF_FSIZE=1, AF_ARG=2, AF_FSTAR=4, AF_HEX=8 } flags; // flags
	PBUF *pPbuf0;						// ptr to 1st PBUF in list
	PBUF *pPbuf;						// ptr to current PBUF (last in list)
} ARGS;

void GetArgs(ARGS *pA);
void WriteData(ARGS *pA);

/*----------------------------------------------------------------------------
;
; Program entry point.  Note we can't use argc/argv -- we get the literal
; command line from the operating system in GetArgs().  This is because we
; need double quotes; can't have the standard library parsing them away.
;
;---------------------------------------------------------------------------*/

int __cdecl main(void) {
	ARGS a;

	GetArgs(&a);						// process command line
	WriteData(&a);						// write generated data to stdout
	return 0;
}

/*----------------------------------------------------------------------------
;
; Display help and exit.
;
;---------------------------------------------------------------------------*/

void HelpExit(void) {					// write help msg to stdout and exit
	fputs(
		"\nstdout ver 2.3 - Generate bytes to stdout.\n\n"
		"  Usage:  stdout {-opts} {file-size{*}}"
				"  hex|{L}\"string\"|@{file} ...\n\n"
		"Example:  stdout 100m ff 0y1010 \"ascii\\n\""
				" @file.txt 0n99 >100meg.tmp\n\n"
		"If 1st arg is string or @file, file-size is total argument size.\n"
		"See stdout.txt or paulhoule.com\\stdout for more options.\n"
	, stdout);
	exit(-1);
}

/*----------------------------------------------------------------------------
;
; Aborts w/error msg to stderr.
; fmt= printf format string, pstr= printf arg (optional, NULL if unused)
;
;---------------------------------------------------------------------------*/

void exiterr(char *fmt, void *pstr) {
	fprintf(stderr, "stdout: ");
	fprintf(stderr, fmt, pstr);
	exit(-1);
}

/*----------------------------------------------------------------------------
;
; Allocate memory; abort w/msg if allocation fails.
;
;---------------------------------------------------------------------------*/

void *mallocerr(size_t bytes) {
	void *pmem = malloc(bytes);
	if (pmem) return pmem;
	exiterr("Memory allocation error\n", NULL);
	return NULL;
}

/*----------------------------------------------------------------------------
;
; 32-bit *pRes = *pRes * mul + i.
; Returns upper 32-bit overflow (0 if none).
;
;---------------------------------------------------------------------------*/

uint32_t m32a(uint32_t *pRes, uint32_t mul, uint32_t i) {
	uint64_t t64 = (uint64_t) *pRes * mul + i;
	*pRes = (uint32_t) t64;
	return (uint32_t) (t64 >> 32);
}

/*----------------------------------------------------------------------------
;
; 64-bit *pRes = *pRes * mul + i.
; Returns upper 32-bit overflow (0 if none).
;
;---------------------------------------------------------------------------*/

uint32_t m64a(uint64_t *pRes, uint32_t mul, uint32_t i) {
	i = m32a((uint32_t *) pRes, mul, i);
	return m32a((uint32_t *) pRes + 1, mul, i);
}

/*----------------------------------------------------------------------------
;
; Convert ascii letter to lowercase (result underfined if c not ascii).
;
;---------------------------------------------------------------------------*/

#define lower(c) ((c) | 0x20)

/*----------------------------------------------------------------------------
;
; Converts character to binary digit of any base (0..35).
; Result is >35 if input char not ascii 0-9,A-Z,a-z.
;
; Function is static so compiler can use a non-standard calling convention
; to minimize register impact without the codespace overhead of inlining.
; 32-bit MSVC ver 7 thru 10: incoming arg passed in eax, and function only
; uses eax -- thus callers know their ecx/edx registers are untouched.
;
;---------------------------------------------------------------------------*/

static unsigned DigToBin(unsigned c) {
	unsigned d = c - '0';
	if (d > 9) {
		d = (d - ('A'-'0') & ~0x20) + 10;
	}
	return d;
}

/*----------------------------------------------------------------------------
;
; Convert string to unsigned 64-bit.
;
; Optional prefix 0x,0n,0t,0y (MSFT MASM convention) forces base 16,10,8,2.
;
; Optional suffix kmgtpe forces decimal kilo(x1000), mega, giga, tera, etc.
; Add i to suffix for power-of-two scaling (eg ki=x1024).
;
; Returns ptr to character stopping conversion.
;
;---------------------------------------------------------------------------*/

wchar_t *wstr2ull(
	uint64_t *pRes,						// &result, max 18446744073709551615
	wchar_t *pC,						// ptr to src
	unsigned base						// non-zero default base (eg 10, 16)
) {
	*pRes = 0;

	if (*pC == '0') {					// if potential 0x, 0y, etc..
		uint8_t *pb = "\x10x\xan\x8t\x2y";
		unsigned i;
		uint32_t c = lower(pC[1]);
		do { i = *pb++; } while (i && *pb++ != c);
		if (i) {
			if (DigToBin(pC[2]) < i) {	// valid digit must follow
				base = i, pC += 2;		// apply base, advance over prefix
			}
		}
	}

	do {								// process digits
		unsigned i = DigToBin(*pC);
		if (i >= base) {				// if bad digit, check for postfix
			uint32_t c = lower(*pC);
			uint32_t j;
			i = 0;
			do {						// this loop is to generate i
				j = "kmgtpe"[i++];
			} while (j && j != c);
			if (!j) break;
			j = 125;					// j will be 1000/8 or 1024/8
			if (pC[1] == 'i') j += 3;
			while (!m64a(pRes, j << 3, 0)) {
				if (--i) continue;
				pC = pC + 1 + (j >> 7);
				break;
			}
			break;
		}
		if (m64a(pRes, (uint32_t) base, (uint32_t) i)) break;
	} while (*++pC);
	return pC;
}

/*----------------------------------------------------------------------------
;
; Allocate a new PBUF and optionally append it to a list.
;
; Normally the default buf[] size is used (see PBUF struct for that).
; A nonzero bsize arg (this is in bytes) overrides the default size.
;
; A total of the data bytes in all (if any) preceding blocks is placed in
; the new totprevbused field.
;
; The sum of totprevbused and bused in the new (last in list) PBUF
; represents the total amount of data bytes in the entire list.
;
; totprevbused and bused count stored data bytes; bbytes is the count of
; bytes allocated to buf[] (if bused == bbytes the buffer is FULL).
;
; No block anywhere in a list must be full (they may be empty).  I.E. do
; not count on bused = bbytes in completed blocks (though it usually is).
;
; pPbuf= ptr to end (pPbuf->pNext == NULL) of a buffer list,
;		or NULL to not add new block to a list (eg starting a new list).
; bsize= if non-zero, overrides the size of the allocated buf.
;
;---------------------------------------------------------------------------*/

PBUF *AllocPbuf(PBUF *pPbuf, unsigned bsize) {
	PBUF *pPtemp;

	if (!bsize) bsize = sizeof(pPtemp->buf); // set default
	pPtemp = mallocerr((sizeof(PBUF) - sizeof(pPtemp->buf)) + bsize + 8);

	pPtemp->pNext = NULL;
	pPtemp->totprevbused = 0;
	pPtemp->bbytes = bsize;
	pPtemp->bused = 0;
	if (pPbuf) {
		pPbuf->pNext = pPtemp;
		pPtemp->totprevbused = pPbuf->totprevbused + pPbuf->bused;
	}
	return pPtemp;
}

/*----------------------------------------------------------------------------
;
; Decodes C escape-sequence into (up to) 32-bit result.
;
; The src pointer must point to the char following the leading \.
;
; A ptr to an ESC struct is 1st arg; on input ->cw must be the encoded
; char width in bytes -- 1 (UTF8) or 2 (UTF16).  On output ->res will hold
; the encoded result; ->cw will be ->res size in bytes (1..4).
; If input ->cw is 1, output ->cw will be 1.
; If input ->cw is 2, output ->cw will be either 2 or 4.
;
; Most values are 1 ASCII byte (eg \n, \0), some 2 bytes (\x1ff, \777).
; Only a \U generating a surrogate pair can be 4 bytes.
;
; If success returns advanced source ptr (beyond parsed esc sequence).
; If source ends with a '\0', source ptr is left pointing at the '\0'.
;
; If failure returns NULL.
;
;---------------------------------------------------------------------------*/

typedef struct {						// used for Esc32 input/output
	uint32_t res;						// parsed result char(s)
	uint32_t cw;						// in/out: char/result byte size
} ESC;

wchar_t *Esc32(ESC *pEsc, wchar_t *pT) {
	uint32_t acc;						// result built here

	{
		/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		; pEnc is a encoding table byte-pair list.
		; 1st byte is encoding info, or 0 if end-of-list.
		; 2nd byte is the ASCII char to encode (the x of \x).
		;
		; If encoding info <= 127, it is the encoded ASCII result byte.
		; Otherwise the info byte's bits are as follows:
		;	7..5: (base)   4 if hex (\x, \u, \U), 7 if octal (\0..\7)
		;	4..4: (unused) 0
		;   3..0: (digits) 0xf if \x, 8 if \U, 4 if \u, 3 if octal
		*/

		uint8_t *pEnc = "\3430" "\217x"	// encoding table (to,from pairs)
		  "\nn" "\210U" "\204u" "\rr" "\\\\" "\"\"" "''" "\tt" "\aa" "\bb"
		  "\33e" "\ff" "\vv" "\??" "\3431\3432\3433\3434\3435\3436\3437";
		unsigned c = *pT++;				// char to encode
		do { acc = *pEnc++; } while (acc && *pEnc++ != c); // look up x in /x
		if (!acc) return NULL;			// error if not matched
	}

	if (acc > 127) do {					// if not 1-byte ASCII (\x, \u, etc),
		uint32_t ec = acc;				// encoding info byte
		acc = 0;						// init digit accumulator

		if (ec & 0x40) {				// if \0..\7 (octal),
			pT--;						// unprocess digit
			ec &= 0x7f;					// change ec[7..5] to 3 (from 7)
		}
		do {							// loop to parse octal/hex digits
			uint32_t dig = DigToBin(*pT); // next potential digit
			uint32_t m = -1;			// mask to check digit and overflow
			uint32_t dw = ec >> 5;		// dw= digit bit width (3 or 4)
			m <<= dw;					// only low digit-width bits now zero
			if (dig & m) break;			// break= not a digit
			m >>= dw;					// only high digit-width bits now zero
			if (acc > m) return NULL;	// abort if append would overflow
			acc = acc << dw | dig;		// append digit to accumulator
			*pT++;						// consume digit
		} while ((uint8_t) ec == 0x8e || --ec & 0xf);

		/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		; Check for insufficient-digits errors.
		*/

		if (!(ec & 0x40)) {				// if \x or \u or \U,
			if ((uint8_t) ec != 0x8e) {	// if not \x with digit(s) entered,
				if (ec & 0xf) return NULL; // error if insuff \x,\u,\U digits
			}
		}

		/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		; Handle simple cw = 1 case -- if no overflow, acc is good as-is.
		*/

		if (pEsc->cw == 1) {			// if encoding width (cw) is 1,
			if (acc <= 0xff) break;		// break= fits in a byte, we're done
			return NULL;				// else error
		}

		/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
		; cw = 2 case -- may have \x or \U > limit; also, may need to encode
		; a \U surrogate pair.  Note: we don't disallow any \u,\U <= limit.
		;
		; We could return a 4-byte \x value; however, that's invalid UTF-16
		; so it's flagged as an error.
		*/

		if (!(ec & 0x4f)) {				// if /U or /u,
			int32_t t;
			if (acc > 0x10ffff) return NULL; // error if invalid unicode
			t = (int32_t) acc - 0x10000; // t= acc over 0x10000
			if (t >= 0) {				// if surrogate pair needed,
				acc = (t >> 10) |		// high surrogate data bits
				  ((t & 0x3ff) << 16);	// low  surrogate data bits
				acc |= 0xdc00d800;		// include surrogate header bits
				pEsc->cw <<= 1;			// double res size (2 chars returned)
			}
		}

		if (ec & 0x8 && acc > 0xffff)	// error if \x exceeds 2 bytes
			return NULL;

	} while (0);
	pEsc->res = acc;					// store encoded result

	return pT;
}

/*----------------------------------------------------------------------------
;
; Parses/stores string (eg "foo\n") given ptr to either leading double quote
; or "L" for wide string (a double quote must follow, or error returned).
;
; The string is appended to the pattern buffer.
;
; If successful, returns advanced source ptr (beyond trailing double quote).
; If source ends with a '\0' (no trailing double quote exists), source ptr
; is left pointing to the trailing '\0'.
;
; If failure returns NULL (some bytes may have been added to pattern buffer).
;
;---------------------------------------------------------------------------*/

wchar_t *GetStr(ARGS *pA, wchar_t *pT) {
	ESC esc;
	unsigned cw = 1;					// cw = 1 for ASCII output

	if (*pT == 'L') {					// if wide string,
		if (*++pT != '"') return NULL;
		cw = sizeof(wchar_t);			// cw = sizeof(wchar_t) for wide char
	}

	pT++;								// advance over leading double quote
	while (1) {
		PBUF *pPbuf = pA->pPbuf;
		if (pPbuf->bbytes - pPbuf->bused < sizeof(uint32_t))
			pA->pPbuf = AllocPbuf(pPbuf, 0); // alloc new block

		esc.cw = cw;
		esc.res = *pT;
		if (!esc.res) break;
		pT++;
		if (esc.res == '\\') {			// if start of esc sequence,
			pT = Esc32(&esc, pT);		// encode esc sequence
			if (!pT) break;				// break= invalid esc sequence
		} else {
			if (esc.res == '"') break;
			if (esc.cw == 1 && esc.res >= 256) // if ASCII must fit in a byte
				return NULL;
		}
		pPbuf = pA->pPbuf;
		*(uint32_t *) (pPbuf->buf + pPbuf->bused) = esc.res; // store result
		pPbuf->bused += esc.cw;			// increase output size of result
	}
	return pT;
}

/*----------------------------------------------------------------------------
;
; Return non-zero if arg char is a recognized (valid) argument terminator.
;
;---------------------------------------------------------------------------*/

int Term(uint32_t c) {
	uint32_t t = ' ' << 24 | '\t' << 16 | '"' << 8;	// t= ' ','\t','"','\0'
	do {								// check for terminator
		if (c == (uint8_t) t) break;
	} while (t >>= 8);

	return t;
}

/*----------------------------------------------------------------------------
;
; Parses/stores hex number.
;
; Parsed number is appended to the pattern buffer.
;
; If successful, returns advanced source ptr (beyond end-of-number).
;
; If failure returns NULL (no data stored, but a new buffer may have been
; allocated).
;
;---------------------------------------------------------------------------*/

wchar_t *GetHex(ARGS *pA, wchar_t *pT) {
	uint8_t *pB;
	PBUF *pPbuf = pA->pPbuf;			// output pattern buffer block

	if (pPbuf->bbytes - pPbuf->bused < 8) // if not room for ull, alloc new
		pA->pPbuf = pPbuf = AllocPbuf(pPbuf, 0); // alloc new block

	pB = pPbuf->buf + pPbuf->bused;		// pB= 1st free buffer byte

	{
		wchar_t *pE = wstr2ull((uint64_t *) pB, pT, 16); // parse hex number
		if (pE == pT) return NULL;
		pT = pE;						// advance over parsed number
	}
	{
		uint32_t c = *pT;				// c = terminator character
		unsigned i = 8;
		while (--i && !pB[i]) ;			// remove high order zeros
		i++;
		if (lower(c) == 'l') {			// if forcing length (0..8)
			i = (unsigned) (pT[1] - '0');
			if (i > 8) return NULL;		// if > 8, bad length postfix
			pT += 2;
		}
		pPbuf->bused += i;				// commit the bytes to buffer
		if (!Term(*pT)) return NULL;	// error if bad terminator
	}
	return pT;
}

/*----------------------------------------------------------------------------
;
; Reads file content into pattern buffer.  fh= open file handle, or NULL
; to read from stdin.
;
; Returns zero if success.
; Returns non-zero if failure (some file data may have been added to buffer).
;
;---------------------------------------------------------------------------*/

int ReadAt(ARGS *pA, FILE *fh) {
	size_t lim;

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; lim is an optimization -- the max file bytes we'll read.
	; If we've already buffered output filesize bytes, it makes no sense
	; to read more (perhaps gigabytes more) that won't be used.
	; (Note: careful... could have already buffered > output filesize.)
	*/

	lim = -1;							// assume we'll read max possible
	if (pA->flags & AF_FSIZE			// if total output size specified,
	 && !(pA->flags & AF_FSTAR)			//  and it's not "times pattern size"
	 && pA->fsize < lim) {				//	and it might limit what we'll read
		size_t tot =					// tot= total bytes buffered so far
			pA->pPbuf->totprevbused + pA->pPbuf->bused;
		lim = 0;						// assume no need to buffer more
		if (tot < (size_t) pA->fsize) {	// if haven't hit/exceeded fsize,
			lim = (size_t) pA->fsize - tot; // limit read to what we need
		}
	}

	if (fh == NULL) {					// handle NULL handle special case
		_setmode(_fileno(fh = stdin), _O_BINARY);
	}

	while (1) {							// loop to read file bytes
		size_t brd;						// bytes read (returned from fread)
		size_t brq =					// free bytes in current buffer (0-n)
		  pA->pPbuf->bbytes - pA->pPbuf->bused;
		if (!brq) {						// if buffer has no free space,
			brq = pA->pPbuf->bbytes;	// assume using previous alloc size
			if (brq < 64*1024*1024) brq *= 2; // double size to a point
			pA->pPbuf = AllocPbuf(pA->pPbuf, (unsigned) brq);
		}
		if (brq > lim) brq = lim;		// enforce limit
		brd = fread(pA->pPbuf->buf + pA->pPbuf->bused, 1, brq, fh);
		pA->pPbuf->bused += (unsigned) brd;
		if (!(lim -= brd)) break;		// break= hit limit, we're done
		if (brd != brq) break;			// break= hit eof (or read error)
	}
	return ferror(fh);
}

/*----------------------------------------------------------------------------
;
; Processes @ argument -- this involves opening a file (if filename is
; specified, otherwise stdin is used) and reading the file/stream content
; into the pattern buffer.  Entry *pT must address lead arg '@'.
;
; If successful, returns advanced source ptr (beyond @ argument).
; If argument ends with a '\0' returned ptr points to the '\0'.
;
; If failure returns NULL (some file data may be added to buffer).
;
;---------------------------------------------------------------------------*/

wchar_t *GetAt(ARGS *pA, wchar_t *pT) {
	enum { fn_elem = 260 };
	wchar_t fn[fn_elem];				// used to hold path+filename
	FILE *fh = NULL;					// NULL causes read from stdin
	unsigned notq;						// nz if NOT quoted filename

	pT++;								// advance over lead @
	do {								// do-nothing block (for breaks)
		wchar_t *p = fn;				// ptr used to store filename
		uint32_t c = *pT;
		notq = c ^ '"';					// notq = nz if no leading quote
		if (!notq) pT++;				// if quote there, advance over it
		while (1) {
			c = *pT;
			if (Term(c)) {				// if terminator,
				if (!c) break;			// '\0' terminates in all cases
				if (notq) break;		// terminate if not inside quotes
				if (c == '"') {			// if quote while inside quotes,
					pT++;				// advance over closing quote
					break;				//	and terminate
				}
			}
			if (p == &fn[fn_elem - 1]) return NULL; // abort if buf overflow
			*p++ = (wchar_t) c;
			pT++;
		}
		*p = '\0';						// terminate filename
		if (p != fn) {					// if filename specified, open it
			fh = _wfopen(fn, L"rb");
			if (!fh) { pT = NULL; break; } // break= error opening file
		}
		if (ReadAt(pA, fh)) { pT = NULL; break; } // read data, break= error

	} while (0);

	if (fh) fclose(fh);					// close named file, if one opened
	return pT;
}

/*----------------------------------------------------------------------------
;
; Advances wide char command line string beyond program name/path (argv[0]).
; Returned string may have leading whitespace.
;
; May be double quotes involved; CommandLineToArgvW argv[0] rules are used.
; CommandLineToArgvW rules obtained from http://www.windowsinspired.com/
;	how-a-windows-programs-splits-its-command-line-into-individual-arguments/
;
;---------------------------------------------------------------------------*/

wchar_t *SkipArgv0(wchar_t *pT) {
	uint32_t nq;
	uint32_t c = *pT;					// c= 1st command line character

	if (c > ' ') {						// if argv[0] exists,
		nq = c ^ '"';					// nq= nz if no lead double quote
		if (!nq) *pT++;					// if lead double quote, consume it
		while (c = *pT) {				// while not at end-of-string,
			*pT++;						// swallow character
			if (nq) {					// if not quoted,
				if (c > ' ') continue;	// consume if valid arg char
				break;					// else consume separator, arg done
			}
			if (c == '"') break;		// quoted: stop if hit ending quote
		}
	}
	return pT;
}

/*----------------------------------------------------------------------------
;
; Parses command line: creates the pattern buffer and sets final output
; filesize (specified or as a function of pattern size).
;
;---------------------------------------------------------------------------*/

__declspec(dllimport) wchar_t * __stdcall GetCommandLineW(void);

void GetArgs(ARGS *pA) {
	wchar_t *pCl, *pT;

	pA->flags = 0;						// init ARGS flags
	pA->pPbuf0 = pA->pPbuf = AllocPbuf(NULL, 0); // init pattern buffer list

	pT = SkipArgv0(GetCommandLineW());	// pT= command line after argv[0]

	do {
		uint32_t c = *pT;				// c= next command line character
		pCl = pT;						// pCl= save start of current arg
		if (!c) break;					// break= hit end of command line

		if (c == ' ' || c == '\t') {	// if whitespace,
			pT++;						// swallow 1 whitespace char
			continue;
		}

		if (!(pA->flags & AF_ARG)		// if haven't seen an argument yet
		  && c >= '0' && c <= '9') {	//	and leading decimal digit,
			wchar_t *pE = wstr2ull(&pA->fsize, pT, 10); // parse as decimal
			if (*pE == '*') *pE++, pA->flags |= AF_FSTAR; // handle '*'
			if (!Term(*pE)) break;		// error if bad terminator
			pT = pE;
			pA->flags |= AF_ARG | AF_FSIZE; // flag arg and fsize seen
			continue;
		}

		if (c == '-' && lower((uint32_t) pT[1]) == 'h') {
			pT += 2, pA->flags |= AF_HEX; // handle -h switch
			continue;
		}

		if (c == '"' || c == 'L') {
			pT = GetStr(pA, pT);		// process ascii or wide string
		} else if (c == '@') {
			pT = GetAt(pA, pT);			// process file data
		} else {

			/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
			; Here arg can only be a hex number (by exclusion of all else),
			; but only if a previous arg has been seen.
			*/

			if (!(pA->flags & AF_ARG)) break; // error if no args seen
			pT = GetHex(pA, pT);
		}
		pA->flags |= AF_ARG;			// flag an argument has been seen
	} while (pT);						// loop if no error occurred

	if (*pCl) exiterr("Argument error %ws\n", pCl);
	if (!(pA->flags & AF_ARG)) HelpExit(); // give help if no arguments

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; If pattern buffer is empty add a single 0 to it (this has no effect
	; on output filesize determination).
	;
	; Generate final output filesize (pA->fsize).  It may be fine already,
	; or it may not have been given and we have to use the pattern size, or
	; if "file-size*" used we have to multiply the pattern size into it.
	*/

	{
		PBUF *pPbuf = pA->pPbuf;		// last in pattern buffer block list
		size_t tot =					// total pattern buffer data bytes
			pPbuf->totprevbused + pPbuf->bused;

		if (!tot) {						// if no data in pattern buffer,
			pPbuf->buf[0] = '\0';		// add a zero (the default data value)
			pPbuf->bused++;
		}

		if (!(pA->flags & AF_FSIZE)) pA->fsize = tot; // use tot if no fsize

		if (pA->flags & AF_FSTAR) {		// if "filesize*" specified,

			/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
			; 32-bit MSVC >= ver 7 knows "(uint64_t) tot >> 32" is 0 (since
			; size_t is 32-bit) and will optimize away much of the below.
			*/

			if ((uint64_t) tot >> 32) {	// if tot > 32 bits,
				uint64_t tfsz = pA->fsize; // swap it with fsize
				pA->fsize = tot;
				tot = (size_t) tfsz;
			}
			if ((uint64_t) tot >> 32	// if multiply overflow,
			 || m64a(&pA->fsize, (uint32_t) tot, 0)) {
				exiterr("file-size* too big\n", NULL);
			}
		}
	}
}

/*----------------------------------------------------------------------------
;
; Writes the parsed output pattern to stdout.
;
;---------------------------------------------------------------------------*/

void WriteData(ARGS *pA) {
	PBUF *pPbuf = pA->pPbuf;			// last in pattern buffer block list

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; If just one pattern block that's less than half full, we can replicate
	; data in the block to enable larger fwrite sizes (don't want to fwrite
	; 1 byte a billion times).  We try to get the block close to full.
	*/

	if (!pPbuf->totprevbused) {			// if all data is in the pPbuf block
		uint32_t n = pPbuf->bused;		// n = # bytes copied to extend buf
		while (pPbuf->bused + n <= sizeof(pPbuf->buf)
		 && pPbuf->bused < pA->fsize) {	// replicate while it makes sense to
			memcpy(pPbuf->buf + pPbuf->bused, pPbuf->buf, n);
			pPbuf->bused += n;			// show block data has increased
			if (n < 120) n <<= 1;		// increase copy size to a point
		}
	}

	/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
	; Write data to stdout.  First we link the last pattern data block to
	; the first, creating a circular list.  That way we can just keep moving
	; to the "next" block until all the data we need to generate is written.
	*/

	_setmode(_fileno(stdout), _O_BINARY); // must write in binary mode

	pPbuf->pNext = pA->pPbuf0;			// create circular list
	while (1) {
		uint32_t bused;

		pPbuf = pPbuf->pNext;			// move to next pattern buffer
		bused = pPbuf->bused;			// data bytes in this ->buf
		if (!bused) continue;			// ignore (it's possible!) empty block

		if (bused > pA->fsize) {		// limit to what's left to write
			bused = (uint32_t) pA->fsize;
			if (!bused) break;			// break= no more data to write
		}
		pA->fsize -= bused;				// update what's left to write

		if (pA->flags & AF_HEX) {		// if writing in hex,
			uint8_t *p = pPbuf->buf;	// dump formatted hex bytes to stdout
			do { printf("%02x ", *p++); } while (--bused);
			continue;
		}
		if (bused != fwrite(pPbuf->buf, 1, bused, stdout)) {
			exiterr("*** File write error ***\n", NULL);
		}
	}
}
